#!/usr/bin/env python3

"""emod-api spatial report module. Exposes SpatialReport and SpatialNode objects."""

import numpy as np


class SpatialNode(object):
    def __init__(self, node_id: int, data):

        self._id = node_id
        self._data = data

        return

    @property
    def id(self) -> int:
        return self._id

    @property
    def data(self):
        """Time series data for this node."""
        return self._data

    # index into node data by time step
    def __getitem__(self, item):
        return self._data[item]

    # index into node data by time step
    def __setitem__(self, key, value):
        self._data[key] = value


NUM_STEPS_INDEX = 0
NUM_NODES_INDEX = 1


class SpatialReport(object):
    def __init__(self, filename: str = None, node_ids: list = None, data=None):

        if isinstance(filename, str):
            self._from_file(filename)
        else:
            self._from_node_ids_and_data(node_ids, data)

        return

    @property
    def data(self) -> np.array:
        return self._data

    @property
    def node_ids(self) -> list:
        return self._node_ids

    @property
    def nodes(self) -> dict:
        return self._nodes

    # index into report by node id
    def __getitem__(self, item) -> SpatialNode:
        return self._nodes[item]

    @property
    def node_count(self) -> int:
        return self.data.shape[NUM_NODES_INDEX]

    @property
    def time_steps(self) -> int:
        return self.data.shape[NUM_STEPS_INDEX]

    def write_file(self, filename: str):

        with open(filename, "wb") as file:
            np.array([self.node_count], dtype=np.uint32).tofile(file)
            np.array([self.time_steps], dtype=np.uint32).tofile(file)
            np.array([self.node_ids], dtype=np.uint32).tofile(file)
            self.data.tofile(file)

        return

    def _from_file(self, filename: str):
        """
        Read binary spatial report file.
        #nodes,
        #time steps,
        node ids (#nodes values),
        data (#nodes x #time steps values)
        """
        # File format:
        # number of nodes      - uint32 * 1
        # number of time steps - uint32 * 1
        # node ids             - uint32 * number of nodes
        # data                 - (float32 * number of nodes) * number of time_steps

        with open(filename, "rb") as file:
            num_nodes = np.fromfile(file, dtype=np.uint32, count=1)[0]
            num_time_steps = np.fromfile(file, dtype=np.uint32, count=1)[0]
            node_ids = np.fromfile(file, dtype=np.uint32, count=num_nodes)
            data = np.fromfile(file, dtype=np.float32, count=num_nodes * num_time_steps)

        # let us index data[step, node]
        data = data.reshape((num_time_steps, num_nodes))
        self._from_node_ids_and_data(node_ids, data)

        return

    def _from_node_ids_and_data(self, node_ids: list, data: np.array):

        assert _is_iterable(node_ids), "node_ids must be specified and iterable"
        concrete = list(node_ids)
        assert len(concrete) > 0, "node_ids must not be empty"
        assert all(map(lambda i: _isinteger(i), concrete)), "node_ids must be integers"
        assert len(set(concrete)) == len(concrete), "node_ids must be unique"
        self._node_ids = sorted(concrete)
        assert data.dtype is np.dtype("float32"), "data must be np.float32"
        assert data.shape[1] == len(
            self._node_ids
        ), "data shape must be (#values, #nodes)"
        self._data = data

        self._node_id_to_index_map = {
            node_ids[n]: n for n in range(data.shape[NUM_NODES_INDEX])
        }
        self._nodes = {
            node_id: SpatialNode(node_id, data[:, self._node_id_to_index_map[node_id]])
            for node_id in node_ids
        }

        return


def _is_iterable(obj) -> bool:
    try:
        _ = iter(obj)
        return True
    except TypeError:
        return False


def _isinteger(item) -> bool:
    return isinstance(item, (int, np.integer))
